// Tasks.cpp: implementation of the CTasks class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#include "TasksSockets.h"
#include "../EmEngine.h"
#include "TaskProcessorSockets.h"
#include "OpCode.h"
#include "../Data/Server.h"
#include "../../otherfunctions.h"

//////////////////////////////////////////////////////////////////////
// global and sockets task functions
CTask::~CTask()
{
}

bool CTask::OnException(CEmEngine&)
{
	// delete this buggy task
	return true;
}

bool CTask_Tcp_Err::Post(CEmEngine& stEngine, SOCKET hSocket, int nErrorCode)
{
	CTask_Tcp_Err* pTask = new CTask_Tcp_Err;
	if (pTask)
	{
		pTask->m_hSocket = hSocket;
		pTask->m_nError = nErrorCode;
		stEngine.PushToSocketsTP(pTask);
	} 
	else
	{
		ASSERT(FALSE);
		AddDebugLogLine("no memory");
	}

	return NULL != pTask;
}

bool CTask_Tcp_Err::Process(CEmEngine& stEngine)
{
	// check if we still have this client
	CEmClient* pClient = stEngine.m_stSocketsTP.Lookup(m_hSocket);
	if (pClient)
	{
		ASSERT(pClient->m_hSocket == m_hSocket);
		// here we can write to the log that this client has disconnected from us

		stEngine.m_stSocketsTP.KillClient(pClient);
	}

	return true;
}

bool CTask_Tcp_Accepted::Process(CEmEngine& stEngine)
{
	CEmClient* pClient = stEngine.m_stSocketsTP.AllocClient(m_eType);
	ASSERT(pClient);
	if (pClient)
	{
		pClient->m_bFromOutside = true;
		pClient->m_hSocket = m_hSocket;
		pClient->m_bIsConnected = true;
		stEngine.m_stSocketsTP.m_mapClients[m_hSocket] = pClient;

		pClient->OnAccepted(stEngine);

	} else
	{
		ASSERT(FALSE);
		AddDebugLogLine("no memory");
		if (closesocket(m_hSocket))
			AddDebugLogLine("closesocket failed");
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_Tcp_Connected::Process(CEmEngine &stEngine)
{
	CEmClient		*pClient = stEngine.m_stSocketsTP.Lookup(m_hSocket);

	if (pClient != NULL)
	{
		ASSERT(!pClient->m_bIsConnected);
		pClient->m_bIsConnected = true;

		pClient->OnConnected(stEngine);
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_Tcp_Web::Process(CEmEngine& stEngine)
{
	CEmClient_Web* pClient = (CEmClient_Web*) stEngine.m_stSocketsTP.Lookup(m_hSocket);
	if (pClient)
	{
		ASSERT(T_CLIENT_WEB == pClient->GetType());

		  CString strTxt;
		  strTxt.Format("Currently I have %d clients", stEngine.m_stSocketsTP.m_mapClients.size());
		  stEngine.m_stTcp.AllocSend(m_hSocket, (LPCTSTR) strTxt, strTxt.GetLength());
		  stEngine.m_stTcp.AllocDisconnect(m_hSocket);
	}

	return true;
}
// end of global and socket task functions
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_LoginToServer::Process(CEmEngine &rEngine)
{
//	Reset message count
//	m_pServer->m_nMsgCount = 0;
	if (m_pServer == NULL)
	{
	//	Report error...
		return true;
	}
	if (m_iMessage == FD_CONNECT)
	{
	//
	//	We've just connected. Make sure we're prepared to handle disconnect.
		CTask_DisconnectServer		*pOnDisconnectTask = new CTask_DisconnectServer();

		pOnDisconnectTask->m_pServer = m_pServer;
		m_pServer->m_pOnCompletionTask = pOnDisconnectTask;

	//	Disconnect from previous connected server, if any
		if ( rEngine.GetServerSocket() != m_pServer->m_hSocket
		&& rEngine.GetServerSocket() != 0 )
		{
			rEngine.DisconnectFromServer(rEngine.GetServerSocket());
		}

	//	Set connection info
		rEngine.ConnectedTo(m_pServer);
		rEngine.SetConnectionState(SERVER_WAITFORLOGIN);

		COpCode_LOGINREQUEST		stMsg;

		md4cpy(stMsg._UserHash, rEngine.GetUserHash());
		stMsg._ClientID = rEngine.GetClientID();
		stMsg._ClientPort = rEngine.GetPort();
		stMsg._UserName.m_strValue = rEngine.GetUserNick();
		stMsg._UserName.m_bValid = true;
		stMsg._ClientSoftVersion.m_dwValue = EDONKEYVERSION;
		stMsg._ClientSoftVersion.m_bValid = true;
		stMsg._UserPort.m_dwValue = rEngine.GetPort();
		stMsg._UserPort.m_bValid = true;

		rEngine.SendOpCode(m_pServer->m_hSocket, stMsg, m_pServer);
	}
	else
	{
	//	We're not expecting this.
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_DisconnectServer::Process(CEmEngine &rEngine)
{
	if (m_pServer == NULL)
		return true;

	// (!) at any reason, can be even if we tried to connect and failed

	if (m_iMessage == FD_CLOSE)
	{
		// Set connection info
		if (rEngine.GetServerSocket() == m_pServer->m_hSocket)
		{
			rEngine.SetConnectionState(SERVER_DISCONNECTED);
		}
		else
		{
			// removed if, because we can get 'disconnected' even if we didn't
			// succeeded to connect yet, thus GetConnectingServer() won't return
			// correct value
			// if(rEngine.GetConnectingServer() == m_hSocket)
			rEngine.SetConnectionState(SERVER_DISCONNECTED);
			// TODO: mark somewhere that server is dead
		}
	}
	else
	{
	//	We're not expecting this.
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// (!) can be both we connected to him or he connected to us
// sending HELLO opcode.
//	MOREVIT: If I'm understanding this correctly, you're saying that
//		when another client connects to us to say HELLO, we're going
//		a HELLO back. What's the point in that? Once the HELLOANSWER
//		is received, both sides have all the information they need
//		about the other.
bool CTask_SayHelloToPeer::Process(CEmEngine &rEngine)
{
	if (m_pPeer == NULL)
		return true;

	if (m_iMessage == FD_CONNECT && !m_pPeer->m_bFromOutside)
	{
		COpCode_HELLO		stMsg;

		md4cpy(stMsg._UserHash, rEngine.GetUserHash());
		stMsg._ClientAddr._IPAddr = rEngine.GetClientID();
		stMsg._ClientAddr._Port = rEngine.GetPort();
		stMsg._UserName.m_strValue = rEngine.GetUserNick();
		stMsg._UserName.m_bValid = TRUE;
		stMsg._ClientSoftVersion.m_dwValue = EDONKEYVERSION;
		stMsg._ClientSoftVersion.m_bValid = TRUE;
		stMsg._UserPort.m_dwValue = rEngine.GetPort();
		stMsg._UserPort.m_bValid = TRUE;

		if (rEngine.IsConnected())
		{
			stMsg._ServerAddr._IPAddr = rEngine.GetConnectedState()->nAddr;
			stMsg._ServerAddr._Port = rEngine.GetConnectedState()->nPort;
		}
		else
		{
			stMsg._ServerAddr._IPAddr = 0;
			stMsg._ServerAddr._Port = 0;
		}

		rEngine.SendOpCode(m_pPeer->m_hSocket, stMsg, m_pPeer);
	}
	else
	{
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connect
CTask_Connect::CTask_Connect(ULONG nAddr, USHORT nPort, T_CLIENT_TYPE eType, CTcpCompletionTask *pOnCompletionTask/*= NULL*/)
	: m_dwAddr(nAddr), m_uPort(nPort), m_eType(eType), m_pOnCompletionTask(pOnCompletionTask)
{
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CTask_Connect::CTask_Connect(LPCTSTR sAddr, USHORT nPort, T_CLIENT_TYPE eType, CTcpCompletionTask *pOnCompletionTask/*= NULL*/) 
	: m_dwAddr(inet_addr(sAddr)), m_uPort(nPort), m_eType(eType), m_pOnCompletionTask(pOnCompletionTask)
{
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_Connect::Process(CEmEngine &stEngine)
{
	if (m_dwAddr != 0 && m_uPort != 0)
	{
		stEngine.m_stSocketsTP.AllocTcpConnect(m_dwAddr, m_uPort, m_eType, m_pOnCompletionTask);
	}
	else
	{
	//	Connect to any server
	//	For now, we'll hardcode Razorback say? I know there's got to be a macro for this.
		m_dwAddr = 195 * 0x1000 + 245 * 0x100 + 244 * 10 + 243;
		m_uPort = 4661;
		stEngine.m_stSocketsTP.AllocTcpConnect(m_dwAddr, m_uPort, m_eType, m_pOnCompletionTask);
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// disconnect
CTask_KillClient::CTask_KillClient(CEmClient *pClient)
	: m_pClient(pClient)
{
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_KillClient::Process(CEmEngine &stEngine)
{
	stEngine.KillClient(m_pClient);

	return true;
}

//////////////////////////////////////////////////////////////////////
// send list of shared files
CTask_SendSharedList::CTask_SendSharedList()
{
	m_bProcessedUI = false;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendSharedList::Process(CEmEngine& stEngine)
{
	if(!stEngine.IsConnected())
		return true;

	// first collect shared files info
/*	if(!m_bProcessedUI)
	{
		stEngine.PushToUI(this);
		return false;
	}
*/
	// afterwards, send it
/*	COpCode_OFFERFILES stMsg;
	for(int i = 0; i < m_Files.GetCount(); i++)
	{
		COpCode_OFFERFILES::CStruct_Files file;
		md4cpy(file._Hash, m_Files[i]._Hash);
		file._Client._Addr = 0;
		file._Client._Port = 0;
		file._FileName.m_strValue = m_Files[i]._FileName;
		file._FileName.m_bValid = true;
		file._FileSize.m_dwValue = m_Files[i]._FileSize;
		file._FileSize.m_bValid = true;
		stMsg._Files.Add(file);
	}
	stEngine.m_stTcp.AllocSend(stEngine.m_stData.GetConnectedServer(), stMsg);
*/
	return true;
}

//////////////////////////////////////////////////////////////////////
/*bool CTask_SendSharedList::ProcessForUI(CEmEngine& stEngine)
{
	CKnownFile	*pKnownFile;
	CCKey		bufKey;
	for (POSITION pos = g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetStartPosition(); pos != 0; )
	{
		g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);
		if(pKnownFile)
		{
			OFFEREDFILE File;

			md4cpy(File._Hash, pKnownFile->GetFileHash());
			File._Addr = 0;
			File._Port = 0;
			File._FileName = pKnownFile->GetFileName();
			File._FileSize = pKnownFile->GetFileSize();

			m_Files.Add(File);

			pKnownFile->SetOfferedToServer(true);
		}
	}

	m_bProcessedUI = true;

	stEngine.m_stSocketsTP.Push(this);

	return false;
}
*/