//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "UploadQueue.h"
#include "packets.h"
#include "emule.h"
#include "knownfile.h"
#ifdef OLD_SOCKETS_ENABLED
#include "ListenSocket.h"
#endif //OLD_SOCKETS_ENABLED
#include "ini2.h"
#include "math.h"
#include "Smtp.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

static uint32	g_iCounter;

VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg,UINT_PTR idEvent,DWORD dwTime);
UINT SaveSourcesThread(HANDLE lpArg);

//	TODO rewrite the whole networkcode, use overlapped sockets

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUploadQueue::CUploadQueue(CPreferences* in_prefs) : waitinglist(100)
{
	EMULE_TRY

	m_pAppPrefs = in_prefs;
	m_hTimer = SetTimer(NULL,141,100,TimerProc);	// KuSh: in msdn : "If the hWnd parameter is NULL, this parameter is ignored", so 141 is unused
	if (!m_hTimer)
		AddDebugLogLine(RGB_LOG_ERROR + _T("%s: Fatal Error, failed to create Timer"), __FUNCTION__);
	m_iEstMaxDataRate = 2000;
	m_dwDataRate = 0;
	m_iBannedCount = 0;
	g_iCounter=0;

//	Extended Upload stats
	m_iULSessionSuccessful = 0;
	m_iULSessionSuccessfulFullChunk = 0;
	for (int i=0; i<ETS_TERMINATOR; i++)
	{
		m_iULSessionSuccessfulPartChunk[i] = 0;
		m_iULSessionFailed[i] = 0;
	}
	m_iTotalUploadTime = 0;
	m_dwSumUpDataOverheadInDeque = 0;
	m_nUpDataRateMSOverhead = 0;
	m_nUpDataRateOverhead = 0;
	m_nUpDataOverheadSourceExchange = 0;
	m_nUpDataOverheadFileRequest = 0;
	m_nUpDataOverheadOther = 0;
	m_nUpDataOverheadServer = 0;
	m_nUpDataOverheadSourceExchangePackets = 0;
	m_nUpDataOverheadFileRequestPackets = 0;
	m_nUpDataOverheadOtherPackets = 0;
	m_nUpDataOverheadServerPackets = 0;

	for (int i = 0;i != 35;i++)
		activeClients_list.AddHead((int)0);

	m_MaxActiveClients = 0;

	m_iLeftOverBandwidth = 0;

	m_guessedMaxBandwidth = 300;
	m_guessedMaxLANBandwidth = 3000;

	m_lastGaveDataTick = ::GetTickCount();
	m_dwLastScheduledBackupTick = ::GetTickCount();

	m_pSaveThread = NULL;
//	Create timer to save known sources every 30 minutes
	m_hSaveSTimer	=	::CreateWaitableTimer(NULL, FALSE, NULL);
//	Create event responsible for thread termination
	m_hQuitEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
	if (m_hSaveSTimer && m_hQuitEvent)
	{
	//	Create lower priority thread that will do the "job"
		m_pSaveThread = AfxBeginThread(SaveSourcesThread, reinterpret_cast<LPVOID>(this),
										THREAD_PRIORITY_BELOW_NORMAL + g_eMuleApp.m_pGlobPrefs->GetMainProcessPriority(),
										CREATE_SUSPENDED);
		if (m_pSaveThread != NULL)
		{
		//	Disable self termination, because we plan to wait for thread termination
			m_pSaveThread->m_bAutoDelete = FALSE;
			m_pSaveThread->ResumeThread();
		}
		else
			AddDebugLogLine(RGB_LOG_ERROR + _T("Error creating SourcesSaveThread"));
	}
	else
	{
		AddDebugLogLine(RGB_LOG_ERROR + _T("Error creating SourcesSaveThread"));
	}

	m_strUploadLogFilePath.Format(_T("%supload.log"), g_eMuleApp.m_pGlobPrefs->GetAppDir());

	EMULE_CATCH2
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::AddClientToUploadQueue(CUpDownClient* pClient)
{
	EMULE_TRY
#ifdef OLD_SOCKETS_ENABLED
	if (g_eMuleApp.m_pListenSocket->TooManySockets())
		return;
#endif //OLD_SOCKETS_ENABLED
	POSITION toadd = 0;
	uint32	bestscore = 0;
	CUpDownClient* pNewSource;

// Select next client or use given client
	if (!pClient)
	{
		POSITION pos1, pos2;
		CUpDownClient* pSource;

		for (pos1 = waitinglist.GetHeadPosition();( pos2 = pos1 ) != NULL;)
		{
			pSource  = waitinglist.GetNext(pos1);

		// Clear dead clients
			ASSERT(pSource->GetLastUpRequest());

		// Remove "?" from queue
			if ( (::GetTickCount() - pSource->GetLastUpRequest() > MAX_PURGEQUEUETIME)
				|| !g_eMuleApp.m_pSharedFilesList->GetFileByID(pSource->m_reqFileHash) )
			{
				RemoveFromWaitingQueue(pos2,true);
#ifdef OLD_SOCKETS_ENABLED
				if (!pSource->m_pRequestSocket)
				{
					if (pSource->Disconnected())
						pSource = NULL;
				}
#endif //OLD_SOCKETS_ENABLED
				continue;
			}
		// Finished clearing
			else if ( (pSource->GetScore(true) > bestscore) && (!pSource->IsBanned())  && (!pSource->HasLowID()
#ifdef OLD_SOCKETS_ENABLED
				|| (pSource->m_pRequestSocket && pSource->m_pRequestSocket->IsConnected())
#endif //OLD_SOCKETS_ENABLED
				) )
			{
				bestscore = pSource->GetScore(true);
				toadd = pos2;
			}
		}
		if (!toadd)
			return;

		pNewSource = waitinglist.GetAt(toadd);
		RemoveFromWaitingQueue(toadd, true);
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();
	}
	else
		pNewSource = pClient;

  	if (!g_eMuleApp.m_pClientList->IsValidClient(pNewSource))
	{
		return;
	}

// Never upload to a banned & already downloading client
	if (IsDownloading(pNewSource) || pNewSource->IsBanned(true))
		return;

// Don't accept clients if requested file is in PS_COMPLETING state to prevent unpredictable behaviour
// (the existence of the requested file will be done in AddClientToQueue)
	CKnownFile* m_pReqPartFile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)pNewSource->m_reqFileHash);
	if (m_pReqPartFile->IsPartFile() && ((CPartFile*)m_pReqPartFile)->GetStatus() == PS_COMPLETING)
		return;

// Tell the client that we are now ready to upload
#ifdef OLD_SOCKETS_ENABLED
	if (!pNewSource->m_pRequestSocket || !pNewSource->m_pRequestSocket->IsConnected())
	{
		pNewSource->SetUploadState(US_CONNECTING);
		if (!pNewSource->TryToConnect(true))
			return;
	}
	else
	{
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ);
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(packet->m_dwSize);
		pNewSource->m_pRequestSocket->SendPacket(packet,true);
		pNewSource->SetUploadState(US_UPLOADING);
		pNewSource->SetLastGotULData();
	}
#endif //OLD_SOCKETS_ENABLED

	pNewSource->SetUpStartTime();
	pNewSource->ResetSessionUp();
	pNewSource->ResetCompressionGain();
	uploadinglist.AddTail(pNewSource);

// Statistic
	if (m_pReqPartFile)
		m_pReqPartFile->statistic.AddAccepted();

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlUploadList.AddClient(pNewSource);

	EMULE_CATCH
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::Process()
{
	EMULE_TRY

	DWORD curTick = ::GetTickCount();
	int timeSinceLastGaveDataMs;

// LANCAST - We count lanclients
	int QtyLANClients = 0;

// After 49 days tick wraps around. I think this still handles it. (Yeah right! As if we wouldn't have crashed by then :))
	timeSinceLastGaveDataMs = curTick - m_lastGaveDataTick;

// Keeps track of how much bandwidth we use this call
    sint32 usedBandwidthThisCall = 0;

// How much data should be given each round?
	if (g_eMuleApp.m_pGlobPrefs->GetMaxUpload() == UNLIMITED)
	{
	// Special case for unlimited upload
		if (CanAcceptNewClient(uploadinglist.GetCount()) && waitinglist.GetCount())
		{
			AddClientToUploadQueue();
		}

		bool clientWasReady = false;

		POSITION pos = uploadinglist.GetHeadPosition();
		while (pos != NULL)
		{
		// Get the client. Note! Also updates activeClientsListPos as a side effect
			CUpDownClient* cur_client = uploadinglist.GetNext(pos);
			bool tempWantsMoreBandwidth;

		// LANCAST - Just Count LAN Clients
			if (cur_client->IsOnLAN())
				QtyLANClients++;
			else
				usedBandwidthThisCall += cur_client->SendBlockData(m_guessedMaxBandwidth-usedBandwidthThisCall, tempWantsMoreBandwidth);

			if (tempWantsMoreBandwidth == true)
				clientWasReady = true;
		}

		if (clientWasReady == false)
		{
			if (m_guessedMaxBandwidth > 300)
				m_guessedMaxBandwidth -= 200;
			else
				m_guessedMaxBandwidth = 100;
		}
		else
		{
			m_guessedMaxBandwidth += 200;
		}

		m_MaxActiveClients = uploadinglist.GetCount();
	}
	else
	{
	// Normal case for limited upload.
		uint32		tempMaxActiveClients = 0;
		POSITION	activeClientsListPos = activeClients_list.GetHeadPosition();

		while (activeClientsListPos != NULL)
		{
			uint32 activeClientsSnapshot = activeClients_list.GetNext(activeClientsListPos);

			if (activeClientsSnapshot > tempMaxActiveClients)
				tempMaxActiveClients = activeClientsSnapshot;
		}

		m_MaxActiveClients = tempMaxActiveClients;
		sint32 realMaxULRatems = (g_eMuleApp.m_pGlobPrefs->GetMaxUpload()*1024*timeSinceLastGaveDataMs)/10000;

#define MINNUMBEROFTRICKLEUPLOADS 2
	// How many slots should be open? Trickle slots included (at least 2 trickles, 30% of total, and no slot is expected to use more than UPLOAD_CLIENT_DATARATE*2 in average, whichever number is largest)
		sint32 wantedNumberOfTotalUploads = max((uint32)(g_eMuleApp.m_pGlobPrefs->GetMaxUpload()*1024/(UPLOAD_CLIENT_DATARATE*20)),
													m_MaxActiveClients + MINNUMBEROFTRICKLEUPLOADS);
		wantedNumberOfTotalUploads = max(wantedNumberOfTotalUploads, m_MaxActiveClients*1.3);

		POSITION			lastpos = uploadinglist.GetTailPosition();
		CUpDownClient	   *lastClient = NULL;

		if (lastpos != NULL)
		{
			lastClient = uploadinglist.GetAt(lastpos);
		}

#define MINWAITBEFOREOPENANOTHERSLOTMS 1000

	// Add or remove connections as needed. Mostly, this is controlled by the number of
	// fully active uploads.
		if (CanAcceptNewClient(uploadinglist.GetCount()) 
			&& waitinglist.GetCount() > 0
			&& wantedNumberOfTotalUploads > uploadinglist.GetCount() 
			&& (lastClient == NULL || lastClient->GetUpStartTimeDelay() > MINWAITBEFOREOPENANOTHERSLOTMS))
		{
		//	There's not enough open uploads. Open more another one.
			AddClientToUploadQueue();
		}
		else if (uploadinglist.GetCount() > MIN_UP_CLIENTS_ALLOWED 
			&&  CanAcceptNewClient(uploadinglist.GetCount()-1) == false)
		{
		// There's to many open uploads (propably due to the user changing
		// the upload limit to a lower value). Remove the last opened upload and put
		// it back on the waitinglist. When it is put back, it get
		// to keep its waiting time. This means it is likely to soon be
		// choosen for upload again.
			if (!lastClient->IsOnLAN() && lastClient->GetUpStartTimeDelay() > 3*1000)
			{
			// Remove from upload list.
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(lastClient, ETS_TIMEOUT, true);

			// add to queue again. This is almost the same as in method
			//    g_eMuleApp.m_pUploadQueue->AddClientToQueue(lastClient,true);
				waitinglist.AddTail(lastClient);
				lastClient->SetUploadState(US_ONUPLOADQUEUE);
				lastClient->SendRankingInfo();
				g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.AddClient(lastClient);
				g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();
			}
		}

	// There's some saved bandwidth that we want to try to spend this round.
	// curMaxULRatems keeps track of how much data we want to spend.
		sint32 curMaxULRatems = realMaxULRatems + m_iLeftOverBandwidth;
		sint32 allowedBandwidthThisCall = min(curMaxULRatems, realMaxULRatems+MAXFRAGSIZE);

	// Which active slot are we giving data each loop
		uint32 slotcounter = 0;

		bool wantsMoreBandWidth = false;

		POSITION ulpos = uploadinglist.GetHeadPosition();
	// The loop that gives the fully activated connections data.
		while (ulpos != NULL && (allowedBandwidthThisCall > usedBandwidthThisCall) && wantsMoreBandWidth == false)
		{
		// Only allow this connection to take what has not already been taken
		// by the other connections above it in the list.
			uint32 allowedThisIteration = allowedBandwidthThisCall-usedBandwidthThisCall;

		// Get the client. Note! Also updates ulpos as a side effect.
		    CUpDownClient* cur_client = uploadinglist.GetNext(ulpos);

		// Give data to the upload, if it wants it. Remember how much data we have given this call.
		// The connection wont take any data if it hasn't emptied its buffer since we last gave it data.
		// variable wantsMoreBandWidth is updated by this method (it is a ref bool&)

		// LANCAST - Increment LanCount
			if (cur_client->IsOnLAN())
				QtyLANClients++;
			else
				usedBandwidthThisCall += cur_client->SendBlockData(allowedThisIteration, wantsMoreBandWidth);

			slotcounter++;
		}

#define MAXTIMESLOTISALLOWEDTOGOWITHOUTDATAMS 2540
	// Trickle the unneeded uploads (just give them enough to not time out)
	// Theese downloads are kept connected, in a ready-to-go state, just in case
	// one of the fully activated uploads completes/timeouts/ends.
	// As soon as there's a little bandwidth leftover, the first one of these
	// uploads will go to fully activated state
		POSITION trickle_client_pos = uploadinglist.FindIndex(slotcounter);

		while (trickle_client_pos != NULL)
		{
		// Get the client. Note! Also updates trickle_client_pos as a side effect.
			CUpDownClient* cur_client = uploadinglist.GetNext(trickle_client_pos);

			if (cur_client->GetLastGotULData() + MAXTIMESLOTISALLOWEDTOGOWITHOUTDATAMS < curTick)
			{
				// It's more than 3 seconds since this connection got any data.
				// Feed it a mercy package to prevent it from timing out.
				bool tempWantsMoreData = false;

				// LANCAST - Increment LanCount
				if (cur_client->IsOnLAN())
					QtyLANClients++;
				else
					usedBandwidthThisCall += cur_client->SendBlockData(MAXFRAGSIZE, tempWantsMoreData);
			}
		}

	// Since we don't save bandwidth for the trickles above, we may
	// have used to much bandwidth.
	// We may also have used to little bandwidth this round, in that
	// case it is saved to the next round.
		m_iLeftOverBandwidth = curMaxULRatems-usedBandwidthThisCall;

	// We don't want to long peaks. Don't save to much bandwidth (about a second should be OK)
	// Always make sure we can save enough to get a packet through
		sint32 limitSave = MAXFRAGSIZE; // max(realMaxULRatems*10, MAXFRAGSIZE+realMaxULRatems*4);
		if (m_iLeftOverBandwidth > limitSave)
			m_iLeftOverBandwidth = limitSave;

	// Save number of active clients for statistics
		activeClients_list.RemoveHead();
		activeClients_list.AddTail(slotcounter);
	}

// Save sent bytes for speed calculations
	if (m_averageTickList.empty() || curTick- m_averageTickList.front() >= UL_QUEUE_DATARATE_SAMPLE_TIME)
	{
		m_averageDataRateList.push_front(g_eMuleApp.stat_sessionSentBytes);
		m_averageTickList.push_front(curTick);

	// Remove anything older than 40 seconds
		while (curTick-m_averageTickList.back() > 40000)
		{
			m_averageDataRateList.pop_back();
			m_averageTickList.pop_back();
		}

	//	recalculate the average speed after list update
		if (m_averageDataRateList.size() > 1)
			m_dwDataRate = static_cast<uint32>(static_cast<double>(m_averageDataRateList.front()-m_averageDataRateList.back())*1000.0 / (m_averageTickList.front()-m_averageTickList.back()));
		else
			m_dwDataRate = 0;
	}
// Save time to be able to calculate how long between each call
	m_lastGaveDataTick = curTick;

// LANCAST - If we have more than one LAN user were going to upload to them
// Were going to try uploading the way plus way.......
	if (QtyLANClients > 0)
	{
		bool clientWasReady = false;
		POSITION pos = uploadinglist.GetHeadPosition();

		while (pos != NULL)
		{
		// Get the client. Note! Also updates activeClientsListPos as a side effect.
			CUpDownClient* cur_client = uploadinglist.GetNext(pos);
			bool tempWantsMoreBandwidth = false;

		// LANCAST - Just Count LAN Clients
			if (cur_client->IsOnLAN())
				cur_client->SendBlockData(m_guessedMaxLANBandwidth, tempWantsMoreBandwidth);

			if (tempWantsMoreBandwidth == true)
				clientWasReady = true;
		}

		if (clientWasReady == false)
		{
			if (m_guessedMaxLANBandwidth > 3000)
				m_guessedMaxLANBandwidth -= 2000;
			else
				m_guessedMaxLANBandwidth = 1000;
		}
		else
		{
			m_guessedMaxLANBandwidth += 2000;
		}
	}

#ifdef OLD_SOCKETS_ENABLED
// enkeyDEV(th1) -L2HAC- lowid side, automatic callback phase
	if (g_eMuleApp.m_pServerConnect->IsConnected() && g_eMuleApp.m_pServerConnect->IsLowID())
	{
		for (POSITION pos = waitinglist.GetHeadPosition(); pos != NULL; )
		{
			CUpDownClient* pSource = waitinglist.GetNext(pos);

			if ( pSource->IsL2HACEnabled() && pSource->GetLastL2HACExecution() && pSource->GetL2HACTime()
				 && (curTick - pSource->GetLastL2HACExecution()) > pSource->GetL2HACTime() )
			{
				if ( g_eMuleApp.m_pListenSocket->TooManySockets()
					 && !(pSource->m_pRequestSocket && pSource->m_pRequestSocket->IsConnected()) )
				{
					 pSource->SetLastL2HACExecution(curTick - pSource->GetL2HACTime() + static_cast<uint32>(ROUND((static_cast<double>(rand())/RAND_MAX)*300000.0)));
				}
				else
				{
					pSource->DisableL2HAC();
					if (!pSource->HasLowID() && pSource->GetL2HACTime())
						pSource->TryToConnect();
				}
			}
		}
	}
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUploadQueue::CanAcceptNewClient(uint32 numberOfUploads)
{
	EMULE_TRY

//	Check if we can allow a new client to start downloading from us
	if (numberOfUploads < MIN_UP_CLIENTS_ALLOWED)
		return true;
	else if (numberOfUploads >= MAX_UP_CLIENTS_ALLOWED)
		return false;

//	Now the final check
	if (g_eMuleApp.m_pGlobPrefs->GetMaxUpload() == UNLIMITED)
	{
		if (numberOfUploads < (GetDataRate()/UPLOAD_CLIENT_DATARATE))
			return true;
	}
	else
	{
		if (numberOfUploads < (uint32)(g_eMuleApp.m_pGlobPrefs->GetMaxUpload()*1024/(UPLOAD_LOW_CLIENT_DR*10))+1)
			return true;
	}
//	Nope
	return false;

	EMULE_CATCH

//	Nope
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUploadQueue::~CUploadQueue()
{
	KillTimer(0, m_hTimer);

	if (m_pSaveThread != NULL)
	{
	//	Request thread termination
		SetEvent(m_hQuitEvent);

	//	Wait for thread termination
		if (::WaitForSingleObject(m_pSaveThread->m_hThread, 1000000) == WAIT_TIMEOUT)
		{
			TRACE("Thread known.met thread did not exited on time");
		}
	//	Delete thread object manually, because self termination was disabled
		delete m_pSaveThread;
	}

	if (m_hSaveSTimer != NULL)
		::CloseHandle(m_hSaveSTimer);
	if (m_hQuitEvent != NULL)
		::CloseHandle(m_hQuitEvent);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
POSITION CUploadQueue::GetWaitingClient(CUpDownClient* client)
{
	EMULE_TRY

	return waitinglist.Find(client);

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CUploadQueue::GetWaitingClientByIPAndUDPPort(uint32 dwIP, uint16 dwPort)
{
	EMULE_TRY

	for (POSITION pos = waitinglist.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient* pSource = waitinglist.GetNext(pos);

		if (dwIP == pSource->GetIP() && dwPort == pSource->GetUDPPort())
			return pSource;
	}

	EMULE_CATCH

	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::UpdateBanCount()
{
	EMULE_TRY

	int count=0;

	for (POSITION pos = waitinglist.GetHeadPosition(); pos != NULL;)
	{
		CUpDownClient	   *pSource = waitinglist.GetNext(pos);

		if (pSource->IsBanned())
			count++;
	}

	SetBanCount(count);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::AddClientToWaitingQueue(CUpDownClient* client, bool bIgnoreTimelimit)
{
	EMULE_TRY

#ifdef OLD_SOCKETS_ENABLED
//	If we're connected to the server, and we're LOWID, and the client is on a remote server and is not downloading
//	and is not a friend and the queue is longer than 50...
	if ( g_eMuleApp.m_pServerConnect->IsConnected()
		&& g_eMuleApp.m_pServerConnect->IsLowID()
		&& !g_eMuleApp.m_pServerConnect->IsLocalServer(client->GetServerIP(),client->GetServerPort())
		&& client->GetDownloadState() == DS_NONE
		&& !client->IsFriend()
		&& GetWaitingUserCount() > 50 )
	{
		return;
	}
#endif //OLD_SOCKETS_ENABLED

//	Filtering invalid eMule clients
	if ( client->GetMuleVersion() == 0 && client->GetVersion() == 0
		&& (client->GetClientSoft() == SO_EMULE || client->GetClientSoft() == SO_PLUS || client->GetClientSoft() == SO_OLDEMULE) )
	{
		return;
	}

//	L2HAC- lowid side
	client->SetLastL2HACExecution();
	if (!client->HasLowID() && client->GetL2HACTime()) client->EnableL2HAC();

//	LANCAST - We ignore any ban LAN clients may have
	if (!bIgnoreTimelimit && !client->IsOnLAN())
	{
		if (client->IsBanned())
		{
			if (GetTickCount() - client->GetBanTime() > g_eMuleApp.m_pGlobPrefs->BadClientBanTime())
				client->UnBan();
			else
				return;
		}
	}

//	Try to add client to our upload queue
	POSITION			pos1, pos2;
	CUpDownClient		*pSource;
	const uint32		dwCurrentTime = ::GetTickCount();
	const uint32		dwWaitingListCount = static_cast<uint32>(waitinglist.GetCount());
	bool				bIsQueueFull = (dwWaitingListCount >= (g_eMuleApp.m_pGlobPrefs->GetQueueSize() + m_iBannedCount));

	for (pos1 = waitinglist.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		pSource = waitinglist.GetNext(pos1);
 
	//	Check if client already exists
		if (pSource == client)
		{
			client->SendRankingInfo();
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(client);
			return;
		}
		else if (client->Compare(pSource)) // Another client with same ip or hash
		{
			if (client->m_pCredits->GetCurrentIdentState(client->GetIP()) == IS_IDENTIFIED && pSource->m_pCredits->GetCurrentIdentState(pSource->GetIP()) != IS_IDENTIFIED)
			{
				AddDebugLogLine( RGB_LOG_DIMMED + _T("Client '%s' (%s) and '%s' (%s) have the same userhash or IP, removed '%s'"),
								 client->GetUserName(), client->GetClientNameAndVersionString(), 
								 pSource->GetUserName(), pSource->GetClientNameAndVersionString(), pSource->GetUserName() );
				RemoveFromWaitingQueue(pos2,true);
			}
			else if (client->m_pCredits->GetCurrentIdentState(client->GetIP()) != IS_IDENTIFIED && pSource->m_pCredits->GetCurrentIdentState(pSource->GetIP()) == IS_IDENTIFIED)
			{
				AddDebugLogLine( RGB_LOG_DIMMED + _T("Client '%s' (%s) and '%s' (%s) have the same userhash or IP, removed '%s'"),
								 client->GetUserName(), client->GetClientNameAndVersionString(),
								 pSource->GetUserName(), pSource->GetClientNameAndVersionString(), client->GetUserName() );
				return;
			}
			else if (client->m_pCredits->GetCurrentIdentState(client->GetIP()) != IS_IDENTIFIED && pSource->m_pCredits->GetCurrentIdentState(pSource->GetIP()) != IS_IDENTIFIED)
			{
				AddDebugLogLine( RGB_LOG_DIMMED + _T("Client '%s' (%s) and '%s' (%s) have the same userhash or IP, removed both"),
								 client->GetUserName(), client->GetClientNameAndVersionString(),
								 pSource->GetUserName(), pSource->GetClientNameAndVersionString() );
				RemoveFromWaitingQueue(pos2,true);
				return;
			}
			else if (client->m_pCredits->GetCurrentIdentState(client->GetIP()) == IS_IDENTIFIED && pSource->m_pCredits->GetCurrentIdentState(pSource->GetIP()) == IS_IDENTIFIED)
			{
				AddDebugLogLine( RGB_LOG_DIMMED + _T("Identified client '%s' (%s) and '%s' (%s) have the same Userhash or IP, removed '%s'"),
								 client->GetUserName(), client->GetClientNameAndVersionString(),
								 pSource->GetUserName(), pSource->GetClientNameAndVersionString(), pSource->GetUserName());
				RemoveFromWaitingQueue(pos2,true);
			}
#ifdef OLD_SOCKETS_ENABLED
			if (pSource->m_pRequestSocket == NULL)
			{
				if (pSource->Disconnected())
					pSource = NULL;
			}
#endif //OLD_SOCKETS_ENABLED
			return;
		}

	//	If client does not exist in UL queue & the queue is full we will not be able to add new client,
	//	therefore we need to check if one of the client can be already removed from the queue due to timeout
	//	(no response from client during MAX_PURGEQUEUETIME)
		if (bIsQueueFull && (dwCurrentTime - pSource->GetLastUpRequest()) > MAX_PURGEQUEUETIME)
		{
			RemoveFromWaitingQueue(pos2,true);
#ifdef OLD_SOCKETS_ENABLED
			if (!pSource->m_pRequestSocket)
			{
				if (pSource->Disconnected())
				{
					pSource = NULL;
				}
			}
#endif //OLD_SOCKETS_ENABLED
		//	Since one place now is free, set full status to false
			bIsQueueFull = false;
		}
	}

//	Use some information from client & add clients server to list.
	if (g_eMuleApp.m_pGlobPrefs->GetAddServersFromConn() && client->GetServerIP() && client->GetServerPort())
	{
	//	If the server isn't already in our server list ...
		if ( (client->GetServerIP() != 0) && !g_eMuleApp.m_pServerList->GetServerByIP(client->GetServerIP(), client->GetServerPort()))
		{
			in_addr		addhost;

			addhost.S_un.S_addr = client->GetServerIP();

			CServer		*pServer = new CServer(client->GetServerPort(), CString(inet_ntoa(addhost)));

			pServer->SetListName(pServer->GetAddress());
			if (!g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.AddServer(pServer, true))
				delete pServer;
		}
	}

	if (client->IsDownloading())
	{
	//	He's already downloading and wants probably only another file
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ);
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
		client->m_pRequestSocket->SendPacket(packet,true);
#endif //OLD_SOCKETS_ENABLED
		client->SetLastGotULData();
		return;
	}

//	If queue is not full we can add a client
	if (!bIsQueueFull)
	{
	//	LANCAST - Lan users bypass the queue
		if (client->IsOnLAN())
			AddClientToUploadQueue(client);
		else
		{
			waitinglist.AddTail(client);
			client->SetWaitStartTime();
			client->SetUploadState(US_ONUPLOADQUEUE);
			client->SendRankingInfo();
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.AddClient(client);
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();
		}
	}

	EMULE_CATCH
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUploadQueue::RemoveFromUploadQueue(CUpDownClient* client, EnumEndTransferSession eReason, bool updatewindow )
{
	if (client == NULL)
		return false;

	EMULE_TRY

	POSITION removed_pos = 0;
	removed_pos = uploadinglist.Find(client);

	if (removed_pos)
	{
		uploadinglist.RemoveAt(removed_pos);

		if (updatewindow)
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlUploadList.RemoveClient(client);

	// Update stats: if client got some bytes from us the session will be succesfull, otherwise failed
		if (client->GetTransferredUp())
		{
			m_iULSessionSuccessful++;

			if (client->GetTransferredUp() >= PARTSIZE)
				m_iULSessionSuccessfulFullChunk++;
			else
				m_iULSessionSuccessfulPartChunk[eReason]++;
		}
		else
		{
			m_iULSessionFailed[eReason]++;
		}

	// If we transfer something we need to update a upload time
		if (client->GetTransferredUp())
		{
			m_iTotalUploadTime += client->GetUpStartTimeDelay()/1000;

			if (g_eMuleApp.m_pGlobPrefs->LogUploadToFile())
			{
				FILE *LogFile = _tfopen(m_strUploadLogFilePath, "a");

				if (LogFile != NULL)
				{
					COleDateTime	currentTime(COleDateTime::GetCurrentTime());
					CString			strLogLine;
					CKnownFile		*m_pReqPartFile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)client->m_reqFileHash);
					uint32			iPartNum = client->GetCurrentlyUploadingPart();

					if (!m_pReqPartFile)
						m_pReqPartFile = g_eMuleApp.m_pKnownFilesList->FindKnownFileByID((uchar*)client->m_reqFileHash);

					strLogLine.Format( _T("%s,\"%s\",%s,\"%s\",%u,%u,%u,%s,%s,%s,%s\n"),
										currentTime.Format(),
										client->GetUserName(),
										client->GetClientNameAndVersionString(),
										(m_pReqPartFile) ? m_pReqPartFile->GetFileName() : _T(""),
										client->GetSessionUp(),
										(iPartNum != 0xFFFF) ? iPartNum : 0,
										(client->GetUpStartTimeDelay()/1000),
										YesNoStr(client->IsFriend()),
										YesNoStr(client->IsCommunity()),
										YesNoStr(client->IsOnLAN()),
										(client->GetSessionUp() >= PARTSIZE) ? 
											GetResString(IDS_IDENTOK).MakeLower() : 
											g_eMuleApp.m_pdlgEmule->m_dlgStatistics.GetUpEndReason(eReason) );
					_fputts(strLogLine, LogFile);
					fclose(LogFile);
				}
			}
		}

		client->SetUploadState(US_NONE);
		client->ClearUploadBlockRequests();

		return true;
	}

	EMULE_CATCH

	return false;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUploadQueue::GetAverageUpTime()
{
	if (m_iULSessionSuccessful)
	{
		return m_iTotalUploadTime/m_iULSessionSuccessful;
	}
	return 0;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUploadQueue::RemoveFromWaitingQueue(CUpDownClient* client, bool updatewindow)
{
	EMULE_TRY

	POSITION pos = waitinglist.Find(client);
	if (pos)
	{
		RemoveFromWaitingQueue(pos,updatewindow);

		if (updatewindow)
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();

		return true;
	}

	EMULE_CATCH

	return false;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::RemoveFromWaitingQueue(POSITION pos, bool updatewindow)
{
	EMULE_TRY

	CUpDownClient* todelete = waitinglist.GetAt(pos);
	waitinglist.RemoveAt(pos);

	if (todelete == NULL)
		return;

	if (todelete->IsBanned())
		todelete->UnBan();

	if (updatewindow)
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.RemoveClient(todelete);

	todelete->SetUploadState(US_NONE);

	EMULE_CATCH
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUploadQueue::CheckForTimeOver(CUpDownClient* client)
{
	EMULE_TRY

	uint32 score = client->GetScore(true,true);

	for (POSITION pos = waitinglist.GetHeadPosition(); pos != NULL;)
	{
		if (score < waitinglist.GetNext(pos)->GetScore(true,false))
			return true;
	}

	EMULE_CATCH

	return false;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::DeleteAll()
{
	EMULE_TRY

	waitinglist.RemoveAll();
	uploadinglist.RemoveAll();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CUploadQueue::GetWaitingPosition(CUpDownClient* client)
{
	EMULE_TRY

	if (!IsOnUploadQueue(client))
		return 0;

	uint16 rank = 1;
	uint32 myscore = client->GetScore(false);

	for (POSITION pos = waitinglist.GetHeadPosition(); pos != NULL; )
	{
		if (waitinglist.GetNext(pos)->GetScore(false) > myscore)
			rank++;
	}
	return rank;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::CompUpDataRateOverhead()
{
	EMULE_TRY

//	update the deque & sum
	m_UpDataOverheadDeque.push_back(m_nUpDataRateMSOverhead);
	m_dwSumUpDataOverheadInDeque += m_nUpDataRateMSOverhead;

//	reset the value
	m_nUpDataRateMSOverhead = 0;

//	check deque length
	while (m_UpDataOverheadDeque.size() > 150)
	{
		m_dwSumUpDataOverheadInDeque -= m_UpDataOverheadDeque.front();
		m_UpDataOverheadDeque.pop_front();
	}

//	calculate average value
	if (m_UpDataOverheadDeque.size() > 10)
		m_nUpDataRateOverhead = 10*m_dwSumUpDataOverheadInDeque/m_UpDataOverheadDeque.size();
	else
		m_nUpDataRateOverhead = 0;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
	EMULE_TRY

//	Don't do anything if the app is shutting down - can cause unhandled exceptions
	if (!g_eMuleApp.m_pdlgEmule->IsRunning())
	{
		return;
	}

	g_eMuleApp.m_pUploadQueue->Process();
	g_eMuleApp.m_pDownloadQueue->Process();
	g_eMuleApp.m_pUploadQueue->CompUpDataRateOverhead();
	g_eMuleApp.m_pDownloadQueue->CompDownDataRateOverhead();

	g_iCounter++;

//	Every second (be aware that every divisor of g_iCounter-modulo has to be a multiple of 10 inside)
	if (g_iCounter % 10 == 0)
	{
#ifdef OLD_SOCKETS_ENABLED
		if (g_eMuleApp.m_pServerConnect->IsConnecting())
		{
			g_eMuleApp.m_pServerConnect->TryAnotherConnectionRequest();
		}
#endif

		if (g_eMuleApp.m_pGlobPrefs->IsWatchClipboard4ED2KLinks())
		{
			g_eMuleApp.m_pdlgEmule->m_dlgSearch.SearchClipBoard();
		}

		g_eMuleApp.m_pdlgEmule->m_dlgStatistics.UpdateConnectionsStatus();

#ifdef OLD_SOCKETS_ENABLED
		if (g_eMuleApp.m_pServerConnect->IsConnecting())
		{
			g_eMuleApp.m_pServerConnect->CheckForTimeout();
		}
#endif

	//	Every 2 seconds
		if (g_iCounter % 20 == 0)
		{
			g_eMuleApp.m_pdlgEmule->m_dlgStatistics.UpdateConnectionStats((double)g_eMuleApp.m_pUploadQueue->GetDataRate()/1024, static_cast<double>(g_eMuleApp.m_pDownloadQueue->GetDataRate())/1024.0);
		}

	//	Display graphs
		if (g_eMuleApp.m_pGlobPrefs->GetTrafficOMeterInterval() > 0)
		{
			if (g_iCounter % ((uint32)(g_eMuleApp.m_pGlobPrefs->GetTrafficOMeterInterval()) * 10) == 0)
			{
				g_eMuleApp.m_pdlgEmule->m_dlgStatistics.SetCurrentRate((static_cast<double>(g_eMuleApp.m_pUploadQueue->GetDataRate()))/1024.0,((double)g_eMuleApp.m_pDownloadQueue->GetDataRate())/1024);
			}
		}

	//	Display stats
		if (g_eMuleApp.m_pdlgEmule->m_pdlgActive == &g_eMuleApp.m_pdlgEmule->m_dlgStatistics && g_eMuleApp.m_pdlgEmule->IsWindowVisible())
		{
			if (g_eMuleApp.m_pGlobPrefs->GetStatsInterval() > 0)
			{
				if (g_iCounter % ((uint32)(g_eMuleApp.m_pGlobPrefs->GetStatsInterval()) * 10) == 0)
				{
					g_eMuleApp.m_pdlgEmule->ShowStatistics();
				}
			}
		}

	//	Save rates every second
		g_eMuleApp.m_pdlgEmule->m_dlgStatistics.RecordRate();

#ifdef OLD_SOCKETS_ENABLED
	//	Mobilemule sockets
		g_eMuleApp.m_pMMServer->Process();
#endif

	//	Every 5 seconds
		if (g_iCounter % 50 == 0)
		{
		#ifdef _DEBUG
			if (!AfxCheckMemory())
			{
				g_eMuleApp.AddLogLine(false, RGB_PINK + _T("** ") + RGB_LOG_ERROR + _T("Memory corruption detected") + RGB_PINK + _T(" **"));
			}
		#endif

		#ifdef OLD_SOCKETS_ENABLED
			g_eMuleApp.m_pListenSocket->Process();
		#endif

			g_eMuleApp.OnlineSig();
			g_eMuleApp.m_pdlgEmule->ShowTransferRate();
		}

	//	Every 5 minutes
		if (g_iCounter % 3000 == 0)
		{
		//	This function does NOT update the tree!
			g_eMuleApp.m_pGlobPrefs->SaveStats();
		}

	//	Every 90 seconds
		if (g_iCounter % 900 == 0)
		{
		//	Auto Sources per File 
			if (g_eMuleApp.m_pGlobPrefs->IsAutoSourcesEnabled())
			{
				g_eMuleApp.m_pDownloadQueue->SetAutoSourcesPerFile();
			}
		}

	//	Every minute
		if (g_iCounter % 600 == 0)
		{
		//	Server keepalive
			g_eMuleApp.m_pServerConnect->KeepConnectionAlive();

		//	Save clients.met (internal 18 minute interval)
			g_eMuleApp.m_pClientCreditList->Process();

		//	Save server.met (internal 17 minute interval)
			g_eMuleApp.m_pServerList->Process();

		//	Save emfriends.met (internal 25 minute interval)
			g_eMuleApp.m_pFriendList->Process();

		//	Scheduled Backup
			if (g_eMuleApp.m_pGlobPrefs->IsScheduledBackup() && g_eMuleApp.m_pGlobPrefs->GetScheduledBackupInterval() != 0)
			{
				if (::GetTickCount() > (g_eMuleApp.m_pUploadQueue->m_dwLastScheduledBackupTick) + g_eMuleApp.m_pGlobPrefs->GetScheduledBackupInterval() * 3600000)
				{
					g_eMuleApp.m_pdlgEmule->RunBackupNow(true);
					g_eMuleApp.m_pUploadQueue->m_dwLastScheduledBackupTick = ::GetTickCount();
					CLoggable::AddDebugLogLine(_T("Scheduled backup performed"));
				}
			}

		//	Scheduler Shift check
			if (g_eMuleApp.m_pGlobPrefs->IsSCHEnabled())
			{
				CTime curr_t = CTime::GetCurrentTime();
				uint32 secs = curr_t.GetSecond() + 60*curr_t.GetMinute() + 60*60*curr_t.GetHour();

				if (secs > g_eMuleApp.m_pGlobPrefs->GetSCHShift1()-60 && secs < g_eMuleApp.m_pGlobPrefs->GetSCHShift1()+60)
				{
				//	Switching to shift1 speeds
					g_eMuleApp.m_pUploadQueue->SCHShift1UploadCheck();
					g_eMuleApp.m_pGlobPrefs->SetMaxUpload(g_eMuleApp.m_pGlobPrefs->GetSCHShift1Upload());
					g_eMuleApp.m_pGlobPrefs->SetMaxDownload(g_eMuleApp.m_pGlobPrefs->GetSCHShift1Download());
					g_eMuleApp.m_pGlobPrefs->SetMaxConnections(g_eMuleApp.m_pGlobPrefs->GetSCHShift1conn());
					g_eMuleApp.m_pGlobPrefs->SetMaxDownloadConperFive(g_eMuleApp.m_pGlobPrefs->GetSCHShift15sec());
				//	Sending message for scheduler shift 1
					CString MessageText;
					MessageText.Format( _T("SCHEDULER: switching to Shift 1 (Max Upload:%.1f Max Download:%.1f Max Connections:%i Max In 5 secs:%i)"),
										static_cast<double>(g_eMuleApp.m_pGlobPrefs->GetSCHShift1Upload()) / 10.0,
										static_cast<double>(g_eMuleApp.m_pGlobPrefs->GetSCHShift1Download()) / 10.0,
										g_eMuleApp.m_pGlobPrefs->GetSCHShift1conn(),
										g_eMuleApp.m_pGlobPrefs->GetSCHShift15sec() );
					g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetUseSchedulerNotifier(), g_eMuleApp.m_pGlobPrefs->IsSMTPInfoEnabled());
					g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_SCHEDULER, false, g_eMuleApp.m_pGlobPrefs->GetUseSchedulerNotifier());
					CLoggable::AddDebugLogLine(MessageText);
				}
				else if (secs > g_eMuleApp.m_pGlobPrefs->GetSCHShift2()-60 && secs < g_eMuleApp.m_pGlobPrefs->GetSCHShift2()+60)
				{
					int dayOfWeek = curr_t.GetDayOfWeek();
					if ( (dayOfWeek==2 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptMon())
						 || (dayOfWeek==3 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptTue())
						 || (dayOfWeek==4 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptWed())
						 || (dayOfWeek==5 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptThu())
						 || (dayOfWeek==6 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptFri())
						 || (dayOfWeek==7 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptSat())
						 || (dayOfWeek==1 && g_eMuleApp.m_pGlobPrefs->IsSCHExceptSun()) )
	               	{
	                   	CLoggable::AddDebugLogLine(_T("SCHEDULER: day excepted!"));
					}
					else
					{
					//	Switching to shift2 speeds
						g_eMuleApp.m_pUploadQueue->SCHShift2UploadCheck();
						g_eMuleApp.m_pGlobPrefs->SetMaxUpload(g_eMuleApp.m_pGlobPrefs->GetSCHShift2Upload());
						g_eMuleApp.m_pGlobPrefs->SetMaxDownload(g_eMuleApp.m_pGlobPrefs->GetSCHShift2Download());
						g_eMuleApp.m_pGlobPrefs->SetMaxConnections(g_eMuleApp.m_pGlobPrefs->GetSCHShift2conn());
						g_eMuleApp.m_pGlobPrefs->SetMaxDownloadConperFive(g_eMuleApp.m_pGlobPrefs->GetSCHShift25sec());
					//	Sending message for scheduler shift 2
						CString MessageText;
						MessageText.Format( _T("SCHEDULER: switching to Shift 2 (Max Upload:%.1f Max Download:%.1f Max Connections:%i Max In 5 secs:%i)"),
											static_cast<double>(g_eMuleApp.m_pGlobPrefs->GetSCHShift2Upload()) / 10.0,
											static_cast<double>(g_eMuleApp.m_pGlobPrefs->GetSCHShift2Download()) / 10.0,
											g_eMuleApp.m_pGlobPrefs->GetSCHShift2conn(),
											g_eMuleApp.m_pGlobPrefs->GetSCHShift25sec() );
						g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetUseSchedulerNotifier(), g_eMuleApp.m_pGlobPrefs->IsSMTPInfoEnabled());
						g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_SCHEDULER, false, g_eMuleApp.m_pGlobPrefs->GetUseSchedulerNotifier());
						CLoggable::AddDebugLogLine(MessageText);
					}
				}
			}
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::FindSourcesForFileById(CTypedPtrList<CPtrList, CUpDownClient*>* srclist, const uchar* filehash)
{
	EMULE_TRY

	POSITION pos = uploadinglist.GetHeadPosition();

	while (pos)
	{
		CUpDownClient	   *pPotentialSource = uploadinglist.GetNext(pos);

		if (md4cmp(pPotentialSource->m_reqFileHash, filehash) == 0)
			srclist->AddTail(pPotentialSource);
	}

	pos = waitinglist.GetHeadPosition();

	while (pos)
	{
		CUpDownClient	   *pPotentialSource = waitinglist.GetNext(pos);

		if (md4cmp(pPotentialSource->m_reqFileHash, filehash) == 0)
			srclist->AddTail(pPotentialSource);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
UINT SaveSourcesThread(HANDLE lpArg)
{
	HANDLE hTimer = (reinterpret_cast<CUploadQueue*>(lpArg))->m_hSaveSTimer;
	HANDLE hArr[2] = { hTimer, (reinterpret_cast<CUploadQueue*>(lpArg))->m_hQuitEvent };
	LARGE_INTEGER llDueTime;

//	The first timer timeout (30 minutes)
	llDueTime.QuadPart = (LONGLONG)-30*60*10000000;
//	Set timer to signal every 30 minutes
	::SetWaitableTimer(hTimer, &llDueTime, 30*60*1000, NULL, NULL, TRUE);

	for (;;)
	{
	//	Wait for timer signal or event requesting to quit
		DWORD dwRC = ::WaitForMultipleObjects(sizeof(hArr)/sizeof(hArr[0]), hArr, FALSE, INFINITE);

	//	Request to quit
		if (dwRC != WAIT_OBJECT_0)
			break;
	//	Check to avoid rare crash on exit as m_pKnownFilesList is deleted before m_pUploadQueue
		if (g_eMuleApp.m_pKnownFilesList == NULL)
			break;
	//	Save the information to the drive
		g_eMuleApp.m_pKnownFilesList->Save();
	}

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUploadQueue::SCHShift1UploadCheck()
{
	if (g_eMuleApp.m_pGlobPrefs->GetSCHShift1Upload() != 0)
	{
		g_eMuleApp.m_pGlobPrefs->SetSCHShift1Download(TieUploadDownload(g_eMuleApp.m_pGlobPrefs->GetSCHShift1Upload(), g_eMuleApp.m_pGlobPrefs->GetSCHShift1Download()));
	}
}

void CUploadQueue::SCHShift2UploadCheck()
{
	if (g_eMuleApp.m_pGlobPrefs->GetSCHShift2Upload() != 0)
	{
		g_eMuleApp.m_pGlobPrefs->SetSCHShift2Download(TieUploadDownload(g_eMuleApp.m_pGlobPrefs->GetSCHShift2Upload(), g_eMuleApp.m_pGlobPrefs->GetSCHShift2Download()));
	}
}

bool CUploadQueue::IsFileUploading (CPartFile* check_file)
{
	EMULE_TRY

	CUpDownClient* pSource;

	for(POSITION pos = uploadinglist.GetHeadPosition(); pos != NULL; )
	{
		pSource = uploadinglist.GetNext(pos);

		if (pSource->m_pReqPartFile == check_file)
			return true;
	}

	return false;

	EMULE_CATCH

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
POSITION CUploadQueue::GetHeadPosition(void)
{
	return waitinglist.GetHeadPosition();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CUploadQueue::GetNext(POSITION& pos)
{
	return waitinglist.GetNext(pos);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
